"""Default execution entry point if running the package via python -m."""

import sys

import cagged.cli


def cli():
    """Run caged from script entry point"""
    return cagged.cli.cli()


if __name__ == "__main__":
    sys.exit(cli())



from abc import abstractmethod
from typing import Optional

from cagged.analyzer.metadata.detector import Detector

MESSAGE = "This package has an empty description on PyPi"


class EmptyInfoDetector(Detector):
    """This heuristic detects packages with empty documentation on the package repository.
    Such situation might be the marker of a low quality package."""

    MESSAGE_TEMPLATE = "This package has an empty description on %s"

    def __init__(self):
        super().__init__(
            name="empty_information",
            description="Identify packages with an empty description field",
        )

    @abstractmethod
    def detect(
        self,
        package_info,
        path: Optional[str] = None,
        name: Optional[str] = None,
        version: Optional[str] = None,
    ) -> tuple[bool, str]:
        """
        Uses a package's information from PyPI's JSON API to determine
        if the package has an empty description

        Args:
            package_info (dict): dictionary representation of PyPI's JSON
                output

        Returns:
            bool: True if package description is empty
            @param **kwargs:
        """
        pass



from abc import abstractmethod
from typing import Optional

from cagged.analyzer.metadata.detector import Detector


class IntegrityMismatch(Detector):
    """This package contains files that have been tampered with between the source repository and the package CDN"""

    RULE_NAME = "repository_integrity_mismatch"

    def __init__(self):
        super().__init__(
            name="repository_integrity_mismatch",
            description="Identify packages with a linked GitHub repository where the package has extra unexpected files",
        )

    @abstractmethod
    def detect(
        self,
        package_info,
        path: Optional[str] = None,
        name: Optional[str] = None,
        version: Optional[str] = None,
    ) -> tuple[bool, str]:
        pass



""" Empty Information Detector

Detects if a package contains an empty description
"""
import configparser
import hashlib
import logging
import os
import re
from typing import Optional, Tuple

import pygit2  # type: ignore
import urllib3.util

from cagged.analyzer.metadata.repository_integrity_mismatch import \
    IntegrityMismatch

GH_REPO_REGEX = r"(?:https?://)?(?:www\.)?github\.com/(?:[\w-]+/)(?:[\w-]+)"
GH_REPO_OWNER_REGEX = r"(?:https?://)?(?:www\.)?github\.com/([\w-]+)/([\w-]+)"

log = logging.getLogger("caged")


def extract_owner_and_repo(url) -> Tuple[Optional[str], Optional[str]]:
    match = re.search(GH_REPO_OWNER_REGEX, url)
    if match:
        owner = match.group(1)
        repo = match.group(2)
        return owner, repo
    return None, None


def find_best_github_candidate(all_candidates_and_highlighted_link, name):
    """
    This method goes through multiple URLs and checks which one is the most suitable to be used as GitHub URL for
    the project repository.
    If the repository homepage is a GitHub URL, it is used in priority
    """
    candidates, best_github_candidate = all_candidates_and_highlighted_link
    # if the project url is a GitHub repository, we should follow this as an instruction. Users will click on it
    if best_github_candidate is not None:
        best_github_candidate = best_github_candidate.replace("http://", "https://")
        url = urllib3.util.parse_url(best_github_candidate)
        if url.host == "github.com":
            return best_github_candidate
    clean_candidates = []
    for entry in candidates:
        # let's do some cleanup
        url = urllib3.util.parse_url(entry)
        if url.host != "github.com":
            continue
        if url.scheme == "http":
            entry = entry.replace("http://", "https://")
        clean_candidates.append(entry)
    for entry in clean_candidates:
        if f"/{name.lower()}" in entry.lower():
            return entry
    # solution 1 did not work, let's be a bit more aggressive
    for entry in clean_candidates:
        owner, repo = extract_owner_and_repo(entry)
        if repo is not None and (
            # Idea: replace by if two strings have a Levenshtein distance < X% of string length
            repo.lower() in name.lower()
            or name.lower() in repo.lower()
        ):
            return entry
    return None


def dict_generator(indict, pre=None):
    """
    This generator recursively go through an arbitrary dict
    Each iteration will be an array containing the path of all leaves of the dict
    """
    pre = pre[:] if pre else []
    if isinstance(indict, dict):
        for key, value in indict.items():
            if isinstance(value, dict):
                for d in dict_generator(value, pre + [key]):
                    yield d
            elif isinstance(value, list) or isinstance(value, tuple):
                for v in value:
                    for d in dict_generator(v, pre + [key]):
                        yield d
            else:
                yield pre + [key, value]
    else:
        yield pre + [indict]


def get_file_hash(path):
    with open(path, "rb") as f:
        # Read the contents of the file
        file_contents = f.read()
        # Create a hash object
        hash_object = hashlib.sha256()
        # Feed the file contents to the hash object
        hash_object.update(file_contents)
        # Get the hexadecimal hash value
        return hash_object.hexdigest(), str(file_contents).strip().splitlines()


def _ensure_proper_url(url):
    parsed = urllib3.util.parse_url(url)
    if parsed.scheme is None:
        url = f"https://{url}"
    return url


def find_github_candidates(package_info) -> Tuple[set[str], Optional[str]]:
    infos = package_info["info"]
    homepage = None

    project_urls = infos.get("project_urls", {})
    # In some cases, the "project_urls" key is set, but is set to None
    if project_urls is None:
        return set(), None

    if "Homepage" in project_urls:
        homepage = package_info["info"]["project_urls"]["Homepage"]
    github_urls = set()
    for dict_path in dict_generator(infos):
        leaf = dict_path[-1]
        if type(leaf) is not str:
            continue
        res = re.findall(GH_REPO_REGEX, leaf)
        if len(res) > 0:
            for cd in res:
                github_urls.add(_ensure_proper_url(cd.strip()))
    best = None
    if homepage in github_urls:
        best = _ensure_proper_url(homepage)
    return github_urls, best


EXCLUDED_EXTENSIONS = [".rst", ".md", ".txt"]


def exclude_result(file_name, repo_root, pkg_root):
    """
    This method filters out some results that are known false positives:
    * if the file is a documentation file (based on its extension)
    * if the fil is an setup.cfg file with the egg_info claim present on Pypi and not on GitHub
    """
    for extension in EXCLUDED_EXTENSIONS:
        if file_name.endswith(extension):
            return True
    if file_name.endswith("setup.cfg"):
        repo_cfg = configparser.ConfigParser()
        repo_cfg.read(os.path.join(repo_root, file_name))
        pkg_cfg = configparser.ConfigParser()
        pkg_cfg.read(os.path.join(pkg_root, file_name))
        repo_sections = list(repo_cfg.keys())
        pkg_sections = list(pkg_cfg.keys())
        if "egg_info" in pkg_sections and "egg_info" not in repo_sections:
            return True
    return False


def find_mismatch_for_tag(repo, tag, base_path, repo_path):
    repo.checkout(tag)
    mismatch = []
    for root, dirs, files in os.walk(base_path):
        relative_path = os.path.relpath(root, base_path)
        repo_root = os.path.join(repo_path, relative_path)
        if not os.path.exists(repo_root):
            continue
        repo_files = list(
            filter(
                lambda x: os.path.isfile(os.path.join(repo_root, x)),
                os.listdir(repo_root),
            )
        )
        for file_name in repo_files:
            if file_name not in files:  # ignore files we don't have in the distribution
                continue
            repo_hash, repo_content = get_file_hash(os.path.join(repo_root, file_name))
            pkg_hash, pkg_content = get_file_hash(os.path.join(root, file_name))
            if repo_hash != pkg_hash:
                if exclude_result(file_name, repo_root, root):
                    continue
                res = {
                    "file": os.path.join(relative_path, file_name),
                    "repo_sha256": repo_hash,
                    "pkg_sha256": pkg_hash,
                }
                mismatch.append(res)
    return mismatch


def find_suitable_tags_in_list(tags, version):
    tag_candidates = []
    for tag_name in tags:
        if tag_name.endswith(version):
            tag_candidates.append(tag_name)
    return tag_candidates


def find_suitable_tags(repo, version):
    tags_regex = re.compile("^refs/tags/(.*)")
    tags = []
    for ref in repo.references:
        match = tags_regex.match(ref)
        if match is not None:
            tags.append(match.group(0))

    return find_suitable_tags_in_list(tags, version)


# Note: we should have the GitHub related logic factored out as we will need it when we check for signed commits
class PypiIntegrityMismatchDetector(IntegrityMismatch):
    """
    This heuristic compares source code available on the package source code repository (e.g. GitHub), and source code
    published on PyPI. If a file is on both sides but has a different content, this heuristic will flag the package.

    This helps identify packages whose release artifacts were modified directly on PyPI.

    Current gaps:
    * Does not check for extraneous files in the release artifacts
    * Does not run it parallel, so can be slow for large code bases
    """

    RULE_NAME = "repository_integrity_mismatch"

    def detect(
        self,
        package_info,
        path: Optional[str] = None,
        name: Optional[str] = None,
        version: Optional[str] = None,
    ) -> tuple[bool, str]:
        if name is None:
            raise Exception("Detector needs the name of the package")
        if path is None:
            raise Exception("Detector needs the path of the package")

        log.debug(
            f"Running repository integrity mismatch heuristic on PyPI package {name} version {version}"
        )
        # let's extract a source repository (GitHub only for now) if we can
        github_urls, best_github_candidate = find_github_candidates(package_info)
        if len(github_urls) == 0:
            return False, "Could not find any GitHub url in the project's description"
        # now, let's find the right url

        github_url = find_best_github_candidate(
            (github_urls, best_github_candidate), name
        )

        if github_url is None:
            return (
                False,
                "Could not find a good GitHub url in the project's description",
            )

        log.debug(f"Using GitHub URL {github_url}")
        # ok, now let's try to find the version! (I need to know which version we are scanning)
        if version is None:
            version = package_info["info"]["version"]
        if version is None:
            raise Exception("Could not find suitable version to scan")
        tmp_dir = os.path.dirname(path)
        if tmp_dir is None:
            raise Exception("no current scanning directory")

        repo_path = os.path.join(tmp_dir, "sources", name)
        repo = pygit2.clone_repository(url=github_url, path=repo_path)

        tag_candidates = find_suitable_tags(repo, version)

        if len(tag_candidates) == 0:
            return False, "Could not find any suitable tag in repository"

        target_tag = None
        # TODO: this one is a bit weak. let's find something stronger - maybe use the closest string?
        for tag in tag_candidates:
            target_tag = tag

        # Idea: parse the code of the package to find the real version - we can grep the project files for
        #  the version, git biscect until we have a file with the same version? will not work if main has not
        #  been bumped yet in version so tags and releases are out only solutions here print(tag_candidates)
        #  Well, that works if we run integrity check for multiple commits

        #  should be good, let's open the sources
        base_dir_name = None
        for entry in os.listdir(path):
            if entry.lower().startswith(
                name.lower().replace("-", "_")
            ) or entry.lower().startswith(name.lower()):
                base_dir_name = entry
        if (
            base_dir_name is None or base_dir_name == "sources"
        ):  # I am not sure how we can get there
            raise Exception("something went wrong when opening the package")
        base_path = os.path.join(path, base_dir_name)

        mismatch = find_mismatch_for_tag(repo, target_tag, base_path, repo_path)
        message = "\n".join(map(lambda x: "* " + x["file"], mismatch))
        return (
            len(mismatch) > 0,
            f"Some files present in the package are different from the ones on GitHub for "
            f"the same version of the package: \n{message}",
        )



import concurrent.futures
import json
import logging
import multiprocessing
import os
import sys
import tempfile
import typing
from abc import abstractmethod
from concurrent.futures import ThreadPoolExecutor

import requests

from cagged.utils.archives import safe_extract

log = logging.getLogger("caged")


def noop(arg: typing.Any) -> None:
    pass


class Scanner:
    def __init__(self) -> None:
        pass

    @abstractmethod
    def scan_local(
        self, path, rules=None, callback: typing.Callable[[dict], None] = noop
    ):
        pass


class ProjectScanner(Scanner):
    def __init__(self, package_scanner):
        super().__init__()
        self.package_scanner = package_scanner

    def _authenticate_by_access_token(self) -> tuple[str, str]:
        """
        Gives Github authentication through access token

        Returns:
            tuple[str, str]: username, personal access token
        """

        user = os.getenv("GIT_USERNAME")
        personal_access_token = os.getenv("GH_TOKEN")
        if not user or not personal_access_token:
            print(
                """WARNING: Please set GIT_USERNAME (Github handle) and GH_TOKEN
                (generate a personal access token in Github settings > developer)
                as environment variables before proceeding."""
            )
            exit(1)
        return (user, personal_access_token)

    def scan_requirements(
        self,
        requirements: str,
        rules=None,
        callback: typing.Callable[[dict], None] = noop,
    ) -> dict:
        """
        Reads the requirements.txt file and scans each possible
        dependency and version

        Args:
            requirements (str): contents of requirements.txt file
            rules: list of rules to apply
            callback: callback to call for each result

        Returns:
            dict: mapping of dependencies to scan results

            ex.
            {
                ....
                <dependency-name>: {
                        issues: ...,
                        results: {
                            ...
                        }
                    },
                ...
            }
        """

        def scan_single_dependency(dependency, version):
            log.debug(f"Scanning {dependency} version {version}")
            result = self.package_scanner.scan_remote(dependency, version, rules)
            return {"dependency": dependency, "version": version, "result": result}

        dependencies = self.parse_requirements(requirements)

        num_workers = multiprocessing.cpu_count()
        if os.environ.get("GUARDDOG_PARALLELISM") is not None:
            num_workers = int(os.environ["GUARDDOG_PARALLELISIM"])

        sys.stderr.write(
            f"Scanning using at most {num_workers} parallel worker threads\n"
        )
        sys.stderr.flush()
        with ThreadPoolExecutor(max_workers=num_workers) as pool:
            try:
                futures: typing.List[concurrent.futures.Future] = []
                for dependency, versions in dependencies.items():
                    assert versions is None or len(versions) > 0
                    if versions is None:

                        futures.append(
                            pool.submit(scan_single_dependency, dependency, None)
                        )
                    else:
                        futures.extend(
                            map(
                                lambda version: pool.submit(
                                    scan_single_dependency, dependency, version
                                ),
                                versions,
                            )
                        )

                results = []
                for future in concurrent.futures.as_completed(futures):
                    result = future.result()
                    if callback is not None:
                        callback(result)
                    results.append(result)
            except KeyboardInterrupt:
                log.warning("Received keyboard interrupt, cancelling scan\n")
                pool.shutdown(wait=False, cancel_futures=True)

        return results

    def scan_remote(self, url: str, branch: str, requirements_name: str) -> dict:
        """
        Scans remote requirements.txt file

        Args:
            url (str): url of the Github repo
            branch (str): branch containing requirements.txt
            requirements_name (str, optional): name of requirements file.
                Defaults to "requirements.txt".

        Returns:
            dict: mapping of dependencies to scan results

            ex.
            {
                ....
                <dependency-name>: {
                        issues: ...,
                        results: {
                            ...
                        }
                    },
                ...
            }
        """

        token = self._authenticate_by_access_token()
        githubusercontent_url = url.replace("github", "raw.githubusercontent")

        req_url = f"{githubusercontent_url}/{branch}/{requirements_name}"
        resp = requests.get(url=req_url, auth=token)

        if resp.status_code == 200:
            return self.scan_requirements(resp.content.decode())
        else:
            sys.stdout.write(
                f"{req_url} does not exist. Check your link or branch name."
            )
            sys.exit(255)

    def scan_local(
        self, path, rules=None, callback: typing.Callable[[dict], None] = noop
    ):
        """
        Scans a local requirements.txt file

        Args:
            path (str): path to requirements.txt file
            rules: list of rules to apply
            callback: callback to call for each result

        Returns:
            dict: mapping of dependencies to scan results

            ex.
            {
                ....
                <dependency-name>: {
                        issues: ...,
                        results: {
                            ...
                        }
                    },
                ...
            }
        """

        try:
            with open(path, "r") as f:
                return self.scan_requirements(f.read(), rules, callback)
        except Exception as e:
            sys.stdout.write(f"Received {e}")
            sys.exit(255)

    @abstractmethod
    def parse_requirements(self, param: str) -> dict[str, set[str]]:
        pass


class PackageScanner(Scanner):
    """
    Scans package for attack vectors based on source code and metadata rules

    Attributes:
        analyzer (Analyzer): Analyzer for source code and metadata rules
    """

    def __init__(self, analyzer):
        super().__init__()
        self.analyzer = analyzer

    def scan_local(
        self, path, rules=None, callback: typing.Callable[[dict], None] = noop
    ) -> (dict, str):
        """
        Scans local package

        Args:
            path (str): path to package
            rules (set, optional): Set of rule names to use. Defaults to all rules.

        Raises:
            Exception: Analyzer exception

        Returns:
            dict: Analyzer output with rules to results mapping
            rules: rules to apply
            callback: callback to call for each result
        """

        if rules is not None:
            rules = set(rules)

        if os.path.exists(path):
            if (
                path.endswith(".tar.gz")
                or path.endswith(".tgz")
                or path.endswith(".zip")
                or path.endswith(".whl")
            ):
                with tempfile.TemporaryDirectory() as tmpdirname:
                    safe_extract(path, tmpdirname)
                    return (
                        self.analyzer.analyze_sourcecode(tmpdirname, rules=rules),
                        path,
                    )
            elif os.path.isdir(path):
                return (
                    self.analyzer.analyze_sourcecode(path, rules=rules),
                    path,
                )
            else:
                raise Exception(
                    f"Path {path} is not a directory nor an archive type supported by caged."
                )
        raise Exception(f"Path {path} does not exist.")

    @abstractmethod
    def download_and_get_package_info(
        self, directory: str, package_name: str, version=None
    ) -> typing.Tuple[dict, str, str]:
        raise NotImplementedError("download_and_get_package_info is not implemented")

    def _scan_remote(
        self, name, base_dir, version=None, rules=None, write_package_info=False
    ):
        directory = os.path.join(os.path.dirname(os.path.abspath(__file__)), base_dir)

        package_path = None
        package_info = None
        try:
            package_info, zip_path, package_path = self.download_and_get_package_info(
                directory, name, version
            )
        except Exception as e:
            log.debug("Unable to download package, ignoring: " + str(e))
            return {"issues": 0, "errors": {"download-package": str(e)}}

        results = self.analyzer.analyze(
            package_path, package_info, rules, name, version
        )
        if write_package_info:
            suffix = f"{name}-{version}" if version is not None else name
            with open(
                os.path.join(results["path"], f"package_info-{suffix}.json"), "w"
            ) as file:
                file.write(json.dumps(package_info))

        return results, zip_path

    def scan_remote(
        self, name, version=None, rules=None, base_dir=None, write_package_info=False
    ):
        """
        Scans a remote package

        Args:
            * `name` (str): name of the package on PyPI
            * `version` (str, optional): version of package (ex. 0.0.1). If not specified, the latest version is assumed
            * `rules` (set, optional): Set of rule names to use. Defaults to all rules.
            * `base_dir` (str, optional): directory to use to download package to. If not specified, a temporary folder
            is created and cleaned up automatically. If not specified, the provided directory is not removed after the
            scan.
            * `write_package_info` (bool, default False): if set to true, the result of the PyPI metadata API is written
             to a json file

        Raises:
            Exception: Analyzer exception

        Returns:
            dict: Analyzer output with rules to results mapping
        """
        if base_dir is not None:
            return self._scan_remote(name, base_dir, version, rules, write_package_info)

        tmpdirname = tempfile.mkdtemp()

        return self._scan_remote(name, tmpdirname, version, rules, write_package_info)

    def download_compressed(self, url, archive_path, target_path):
        """Downloads a compressed file and extracts it

        Args:
            url (str): download link
            archive_path (str): path to download compressed file
            target_path (str): path to unzip compressed file
        """

        log.debug(f"Downloading package archive from {url} into {target_path}")
        response = requests.get(url, stream=True)

        with open(archive_path, "wb") as f:
            f.write(response.raw.read())

        try:
            safe_extract(archive_path, target_path)
            log.debug(f"Successfully extracted files to {target_path}")
        finally:
            log.debug(f"Removing temporary archive file {archive_path}")



import hashlib
import json
import os.path

import yaml

from cagged.analyzer.metadata import get_metadata_detectors
from cagged.ecosystems import ECOSYSTEM


def build_rules_help_list() -> dict:
    """
    Builds a dict with the names of all available rules and their documentation
    @return: dict[name_of_rule, rule_description]
    """
    rules_documentation = {}
    for ecosystem in ECOSYSTEM:
        rules = get_metadata_detectors(ecosystem)
        for name, instance in rules.items():
            detector_class = instance.__class__.__base__
            rules_documentation[name] = detector_class.__doc__
    dir_path = os.path.dirname(os.path.realpath(__file__))
    semgrep_rules_base_dir = os.path.join(dir_path, "..", "analyzer", "sourcecode")
    for file in os.listdir(semgrep_rules_base_dir):
        if not file.endswith(".yml") and not file.endswith(".yaml"):
            continue
        with open(os.path.join(semgrep_rules_base_dir, file), "r") as fd:
            content = yaml.safe_load(fd)
            for rule in content["rules"]:
                text = rule["description"] if "description" in rule else rule["message"]
                rules_documentation[rule["id"]] = text
    return rules_documentation


def get_sarif_log(runs):
    """
    https://docs.github.com/en/code-security/code-scanning/integrating-with-code-scanning/sarif-support-for-code-scanning#sariflog-object
    """
    return {
        "$schema": "https://json.schemastore.org/sarif-2.1.0.json",
        "version": "2.1.0",
        "runs": runs,
    }


def get_run(results, driver):
    """
    https://docs.github.com/en/code-security/code-scanning/integrating-with-code-scanning/sarif-support-for-code-scanning#run-object
    """
    return {"tool": {"driver": driver}, "results": results}


def get_driver(rules, ecosystem: str):
    """
    https://docs.github.com/en/code-security/code-scanning/integrating-with-code-scanning/sarif-support-for-code-scanning#toolcomponent-object
    """
    return {"name": f"caged-{ecosystem}", "rules": rules}


def get_rule(rule_name: str, rules_documentation) -> dict:
    """
    https://docs.github.com/en/code-security/code-scanning/integrating-with-code-scanning/sarif-support-for-code-scanning#reportingdescriptor-object
    """
    message = (
        rules_documentation[rule_name]
        if rules_documentation[rule_name] is not None
        else ""
    )
    return {
        "id": rule_name,
        "defaultConfiguration": {"level": "warning"},
        "shortDescription": {"text": f"caged rule: {rule_name}"},
        "fullDescription": {"text": message},
        "help": {"text": message, "markdown": message},
        "properties": {"precision": "medium"},
    }


def get_result(rule_name, locations, text, partial_fingerprints):
    """
    https://docs.github.com/en/code-security/code-scanning/integrating-with-code-scanning/sarif-support-for-code-scanning#result-object
    """
    return {
        "ruleId": rule_name,
        "message": {"text": text},
        "locations": locations,
        "partialFingerprints": partial_fingerprints,
    }


def get_location(physical_location):
    """
    https://docs.github.com/en/code-security/code-scanning/integrating-with-code-scanning/sarif-support-for-code-scanning#location-object
    """
    return {"physicalLocation": physical_location}


def get_physical_location(uri, region):
    """
    https://docs.github.com/en/code-security/code-scanning/integrating-with-code-scanning/sarif-support-for-code-scanning#physicallocation-object
    """
    return {"artifactLocation": {"uri": uri}, "region": region}


def get_region(package_raw: str, package: str) -> dict:
    start_line = 0
    start_column = 0
    end_column = 0
    for idx, val in enumerate(package_raw.split("\n")):
        if package in val:
            start_line = idx + 1
            start_column = val.index(package) + 1
            end_column = start_column + len(package)

    return {
        "startLine": start_line,
        "endLine": start_line,
        "startColumn": start_column,
        "endColumn": end_column,
    }


def report_verify_sarif(
    package_path: str,
    rule_names: list[str],
    scan_results: list[dict],
    ecosystem: ECOSYSTEM,
) -> str:
    rules_documentation = build_rules_help_list()
    rules = list(map(lambda s: get_rule(s, rules_documentation), rule_names))
    driver = get_driver(rules, ecosystem.value)
    results = []

    with open(package_path, "r") as file:
        package_raw = file.read()

    for entry in scan_results:
        if entry["result"]["issues"] == 0:
            continue

        region = get_region(package_raw, entry["dependency"])
        uri = package_path[2:] if package_path.startswith("./") else package_path
        physical_location = get_physical_location(uri, region)
        location = get_location(physical_location)
        scan_result_details = entry["result"]["results"]
        package = entry["dependency"]
        version = entry["version"]
        for rule_name in scan_result_details.keys():
            if len(scan_result_details[rule_name]) == 0:
                continue
            text = (
                f"On package: {package} version: {version}\n"
                + "\n".join(map(lambda x: x["message"], scan_result_details[rule_name]))
                if type(scan_result_details[rule_name]) == list
                else scan_result_details[rule_name]
            )
            key = f"{rule_name}-{text}"
            partial_fingerprints = {
                f"caged/v1/{rule_name}": hashlib.sha256(key.encode("utf-8")).hexdigest()
            }
            result = get_result(rule_name, [location], text, partial_fingerprints)
            results.append(result)

    runs = get_run(results, driver)
    log = get_sarif_log([runs])
    return json.dumps(log, indent=2)



import logging
import os
import pathlib
import typing
from urllib.parse import urlparse

import requests

from cagged.analyzer.analyzer import Analyzer
from cagged.ecosystems import ECOSYSTEM
from cagged.scanners.scanner import PackageScanner

log = logging.getLogger("caged")


class NPMPackageScanner(PackageScanner):
    def __init__(self) -> None:
        super().__init__(Analyzer(ECOSYSTEM.NPM))

    def download_and_get_package_info(
        self, directory: str, package_name: str, version=None
    ) -> typing.Tuple[dict, str]:
        git_target = None
        if urlparse(package_name).hostname is not None and package_name.endswith(
            ".git"
        ):
            git_target = package_name

        if not package_name.startswith("@") and package_name.count("/") == 1:
            git_target = f"https://github.com/{package_name}.git"

        if git_target is not None:
            raise Exception("Git targets are not yet supported for npm")

        url = f"https://registry.npmjs.org/{package_name}"
        log.debug(f"Downloading NPM package from {url}")
        response = requests.get(url)

        if response.status_code != 200:
            raise Exception(
                "Received status code: " + str(response.status_code) + " from npm"
            )
        data = response.json()
        if "name" not in data:
            raise Exception(f"Error retrieving package: {package_name}")

        version = data["dist-tags"]["latest"] if version is None else version

        details = data["versions"][version]

        tarball_url = details["dist"]["tarball"]
        file_extension = pathlib.Path(tarball_url).suffix
        zippath = os.path.join(
            directory, package_name.replace("/", "-") + file_extension
        )
        unzippedpath = zippath.removesuffix(file_extension)
        self.download_compressed(tarball_url, zippath, unzippedpath)

        return data, unzippedpath



""" Empty Information Detector

Detects if a package contains an empty description
"""
import os.path
from typing import Optional

from cagged.analyzer.metadata.empty_information import EmptyInfoDetector

MESSAGE = "This package has an empty description on PyPi"


class NPMEmptyInfoDetector(EmptyInfoDetector):
    def detect(
        self,
        package_info,
        path: Optional[str] = None,
        name: Optional[str] = None,
        version: Optional[str] = None,
    ) -> tuple[bool, str]:
        if path is None:
            raise TypeError("path must be a string")
        package_path = os.path.join(path, "package")
        content = map(lambda x: x.lower(), os.listdir(package_path))
        return "readme.md" not in content, EmptyInfoDetector.MESSAGE_TEMPLATE % "npm"



from cagged.analyzer.metadata.detector import Detector


class ReleaseZeroDetector(Detector):
    """This heuristic detects if the latest release of this package is version 0."""

    MESSAGE_TEMPLATE = "The package has its latest release version to %s"

    def __init__(self):
        super().__init__(
            name="release_zero",
            description="Identify packages with an release version that's 0.0 or 0.0.0",
        )



import logging
import os
import zipfile

import tarsafe

log = logging.getLogger("caged")


def safe_extract(source_archive: str, target_directory: str) -> None:
    """
    safe_extract safely extracts archives to a target directory.

    This function does not clean up the original archive, and does not create the target directory if it does not exist.

    @param source_archive:      The archive to extract
    @param target_directory:    The directory where to extract the archive to
    @raise ValueError           If the archive type is unsupported
    """
    log.debug(f"Extracting archive {source_archive} to directory {target_directory}")
    if source_archive.endswith(".tar.gz") or source_archive.endswith(".tgz"):
        tarsafe.open(source_archive).extractall(target_directory)
    elif source_archive.endswith(".zip") or source_archive.endswith(".whl"):
        with zipfile.ZipFile(source_archive, "r") as zip:
            for file in zip.namelist():

                zip.extract(file, path=os.path.join(target_directory, file))
    else:
        raise ValueError("unsupported archive extension: " + target_directory)



import json
import os
from typing import Dict

import docker

from cagged.misc import console
from cagged.ml.analyze import analyze_ml


def run_analysis(
    ecosystem: str,
    package: str,
    package_path: str = "",
    dry_run: bool = False,
    fully_offline: bool = False,
) -> bool:
    caged_results_dir = "/tmp/caged"
    results_dir = f"{caged_results_dir}/results"

    docker_opts = ["run"]
    docker_mounts = [
        "-v",
        "/var/lib/containers:/var/lib/containers",
        "-v",
        f"{results_dir}:/results",
    ]
    analysis_image = "gcr.io/ossf-malware-analysis/analysis"
    analysis_args = [
        "analyze",
        "-upload",
        "file:///results/",
        "-ecosystem",
        ecosystem,
        "-package",
        package,
    ]

    package_path = os.path.realpath(os.path.expanduser(package_path))
    mounted_pkg_path = f"/{os.path.basename(package_path)}"
    docker_mounts += ["-v", f"{package_path}:{mounted_pkg_path}"]
    analysis_args += ["-local", mounted_pkg_path]

    if fully_offline:
        docker_opts += ["--network", "none"]
    if dry_run:
        command = {
            "command": f"docker {' '.join(docker_opts)} {' '.join(docker_mounts)} {analysis_image} {' '.join(analysis_args)}"
        }
        print(command)
        return command

    if not os.path.exists(results_dir):
        os.makedirs(results_dir)

    docker_client = docker.from_env()

    container = docker_client.containers.run(
        image=analysis_image,
        command=" ".join(analysis_args),
        detach=True,
        remove=True,
        cgroupns="host",
        privileged=True,
        tty=True,
        volumes={
            results_dir: {"bind": "/results", "mode": "rw"},
            package_path: {"bind": mounted_pkg_path, "mode": "rw"},
        },
    )

    output = container.attach(stdout=True, stream=True, logs=True)
    for line in output:
        parts = line.decode().strip().split(maxsplit=3)
        if len(parts) == 4:
            time, level, file, message = parts
            if level == "INFO":
                console.log(message)

    with open("/tmp/caged/results/results.json") as f:
        analysis_output = json.load(f)

    console.log(f"💬 Analyzing results using Machine Learning...", emoji=True)

    return analyze_ml(analysis_output)



import os
import pathlib

import yaml
from yaml.loader import SafeLoader

from cagged.ecosystems import ECOSYSTEM

current_dir = pathlib.Path(__file__).parent.resolve()
rule_file_names = list(filter(lambda x: x.endswith("yml"), os.listdir(current_dir)))

SOURCECODE_RULES = {
    ECOSYSTEM.PYPI: list(),
    ECOSYSTEM.NPM: list(),
}

for file_name in rule_file_names:
    with open(os.path.join(current_dir, file_name), "r") as fd:
        data = yaml.load(fd, Loader=SafeLoader)
        for rule in data["rules"]:
            for lang in rule["languages"]:
                match lang:
                    case "python":
                        SOURCECODE_RULES[ECOSYSTEM.PYPI].append(rule)
                    case "javascript" | "typescript" | "json":
                        SOURCECODE_RULES[ECOSYSTEM.NPM].append(rule)



import os
import re

import joblib
import pandas as pd
import rich
from rich.table import Table


def vectorize_package_info(package_info):
    analysis = package_info["Analysis"]

    if analysis.get("import") is None:
        return None
    import_analysis = analysis["import"]
    import_analysis_files = import_analysis["Files"]
    import_analysis_sockets = import_analysis["Sockets"]
    import_analysis_commands = import_analysis["Commands"]
    import_analysis_dns = import_analysis["DNS"]

    num_import_sockets = 0
    num_import_web_sockets = 0
    num_import_dns_sockets = 0
    num_import_ftp_sockets = 0
    num_import_zero_sockets = 0
    num_import_other_sockets = 0
    num_import_local_sockets = 0
    if import_analysis_sockets is not None:
        for socket in import_analysis_sockets:
            if socket["Port"] == 80 or socket["Port"] == 443:
                num_import_web_sockets += 1
            elif socket["Port"] == 53:
                num_import_dns_sockets += 1
            elif socket["Port"] == 21:
                num_import_ftp_sockets += 1
            elif socket["Port"] == 0:
                num_import_zero_sockets += 1
            else:
                num_import_other_sockets += 1

            if socket["Address"] == "::1" or socket["Address"] == "127.0.0.1":
                num_import_local_sockets += 1

    num_import_commands = (
        0 if import_analysis_commands is None else len(import_analysis_commands)
    )
    num_import_dns_records = (
        0 if import_analysis_dns is None else len(import_analysis_dns)
    )
    num_import_files = (
        0 if import_analysis_files is None else len(import_analysis_files)
    )
    num_import_read_files = 0
    num_import_write_files = 0
    num_import_delete_files = 0
    if import_analysis_files:
        for file in import_analysis_files:
            if file["Read"]:
                num_import_read_files += 1
            if file["Write"]:
                num_import_write_files += 1
            if file["Delete"]:
                num_import_delete_files += 1

    return {
        "importDNS": num_import_dns_records,
        "importCommands": num_import_commands,
        "importSockets": num_import_sockets,
        "importWebSockets": num_import_web_sockets,
        "importDNSSockets": num_import_dns_sockets,
        "importFTPSockets": num_import_ftp_sockets,
        "importZeroSockets": num_import_zero_sockets,
        "importOtherSockets": num_import_other_sockets,
        "importLocalSockets": num_import_local_sockets,
        "importFiles": num_import_files,
        "importReadFiles": num_import_read_files,
        "importWriteFiles": num_import_write_files,
        "importDeleteFiles": num_import_delete_files,
    }


def analyze_ml(package_info: dict) -> bool:
    """
    Analyze a package using a machine learning model and predict if it is malicious or not.
    :param package_info:
    :return: bool (True if malicious, False if not)
    """
    vectorized = vectorize_package_info(package_info)

    print()
    table = Table(
        *[
            "DNS Records",
            "Commands",
            "Sockets",
            "Web Sockets",
            "DNS Sockets",
            "FTP Sockets",
            "Zero Sockets",
            "Other Sockets",
            "Local Sockets",
            "Files",
            "Read Files",
            "Write Files",
            "Delete Files",
        ],
        title="Source code analysis",
    )

    table.add_row(*[str(value) for value in vectorized.values()])

    rich.print(table)
    print()

    df = pd.DataFrame([vectorized])

    with open(os.path.join(os.getcwd(), "caged", "ml", "model.joblib"), "rb") as r:
        model = joblib.load(r)

    prediction = model.predict(df)

    return prediction[0]



from cagged.analyzer.metadata.npm.empty_information import NPMEmptyInfoDetector
from cagged.analyzer.metadata.npm.potentially_compromised_email_domain import \
    NPMPotentiallyCompromisedEmailDomainDetector
from cagged.analyzer.metadata.npm.release_zero import NPMReleaseZeroDetector
from cagged.analyzer.metadata.npm.typosquatting import NPMTyposquatDetector

NPM_METADATA_RULES = {}

classes = [
    NPMEmptyInfoDetector,
    NPMReleaseZeroDetector,
    NPMPotentiallyCompromisedEmailDomainDetector,
    NPMTyposquatDetector,
]

for detectorClass in classes:
    detectorInstance = detectorClass()  # type: ignore
    NPM_METADATA_RULES[detectorInstance.get_name()] = detectorInstance



""" Empty Information Detector

Detects if a package contains an empty description
"""
import logging
from typing import Optional

from cagged.analyzer.metadata.empty_information import EmptyInfoDetector

MESSAGE = "This package has an empty description on PyPi"

log = logging.getLogger("caged")


class PypiEmptyInfoDetector(EmptyInfoDetector):
    def detect(
        self,
        package_info,
        path: Optional[str] = None,
        name: Optional[str] = None,
        version: Optional[str] = None,
    ) -> tuple[bool, str]:
        log.debug(
            f"Running PyPI empty description heuristic on package {name} version {version}"
        )
        return (
            len(package_info["info"]["description"].strip()) == 0,
            EmptyInfoDetector.MESSAGE_TEMPLATE % "PyPI",
        )



from cagged.analyzer.metadata.detector import Detector
from cagged.analyzer.metadata.npm import NPM_METADATA_RULES
from cagged.analyzer.metadata.pypi import PYPI_METADATA_RULES
from cagged.ecosystems import ECOSYSTEM


def get_metadata_detectors(ecosystem: ECOSYSTEM) -> dict[str, Detector]:
    match (ecosystem):
        case ECOSYSTEM.PYPI:
            return PYPI_METADATA_RULES
        case ECOSYSTEM.NPM:
            return NPM_METADATA_RULES



from abc import abstractmethod
from datetime import datetime
from typing import Optional

import whois

from cagged.analyzer.metadata.detector import Detector


class PotentiallyCompromisedEmailDomainDetector(Detector):
    def __init__(self, ecosystem: str):
        super().__init__(
            name="potentially_compromised_email_domain",
            description="Identify when a package maintainer e-mail domain (and therefore package manager account) "
            "might have been compromised",
        )
        self.ecosystem = ecosystem

    def _get_domain_creation_date(
        self, email_domain
    ) -> tuple[Optional[datetime], bool]:
        """
        Gets the creation date of an email address domain

        Args:
            email_domain (str): domain of email address

        Raises:
            Exception: "Domain {email_domain} does not exist"

        Returns:
            datetime: creation date of email_domain
            bool:     if the domain is currently registered
        """

        try:
            domain_information = whois.whois(email_domain)
        except whois.parser.PywhoisError as e:

            return None, (not str(e).lower().startswith("no match for"))

        if domain_information.creation_date is None:

            return None, True

        creation_dates = domain_information.creation_date

        if type(creation_dates) is list:
            return min(creation_dates), True

        return creation_dates, True

    def detect(
        self,
        package_info,
        path: Optional[str] = None,
        name: Optional[str] = None,
        version: Optional[str] = None,
    ) -> tuple[bool, str]:
        """
        Uses a package's information from PyPI's JSON API to determine
        if the package's email domain might have been compromised

        Args:
            package_info (dict): dictionary representation of PyPI's JSON
                output

        Raises:
            Exception: "Email for {package_info['info']['name']} does not exist."

        Returns:
            bool: True if email domain is compromised
        """

        emails = self.get_email_addresses(package_info)

        if len(emails) == 0:

            return False, "No e-mail found for this package"

        latest_project_release = self.get_project_latest_release_date(package_info)

        has_issues = False
        messages = []
        for email in emails:
            sanitized_email = email.strip().replace(">", "").replace("<", "")
            email_domain = sanitized_email.split("@")[-1]
            domain_creation_date, domain_exists = self._get_domain_creation_date(
                email_domain
            )

            if not domain_exists:
                has_issues = True
                messages.append(
                    f"The maintainer's email ({email}) domain does not exist and can likely be registered "
                    f"by an attacker to compromise the maintainer's {self.ecosystem} account"
                )
            if domain_creation_date is None or latest_project_release is None:
                continue
            if latest_project_release < domain_creation_date:
                has_issues = True
                messages.append(
                    f"The domain name of the maintainer's email address ({email}) was"
                    " re-registered after"
                    " the latest release of this "
                    "package. This can be an indicator that this is a"
                    ""
                    " custom domain that expired, and was leveraged by"
                    " an attacker to compromise the"
                    f" package owner's {self.ecosystem}"
                    " account."
                )
        return has_issues, "\n".join(messages)

    @abstractmethod
    def get_project_latest_release_date(self, package_info):
        pass

    @abstractmethod
    def get_email_addresses(self, package_info):
        pass



import os
from typing import Optional

from cagged.analyzer.metadata.detector import Detector

THRESHOLD = 1


class PypiSinglePythonFileDetector(Detector):
    def __init__(self):
        super().__init__(
            name="single_python_file",
            description="Identify packages that have only a single Python file",
        )

    def detect(
        self,
        package_info,
        path: Optional[str] = None,
        name: Optional[str] = None,
        version: Optional[str] = None,
    ) -> tuple[bool, Optional[str]]:
        if path is None:
            raise ValueError("path is needed to run heuristic " + self.get_name())
        matches = self._has_fewer_than_threshold_python_files(path)
        return matches, f"This package has {THRESHOLD} or fewer Python source files"

    def _has_fewer_than_threshold_python_files(self, path: str) -> bool:
        # We could use something like the below instead:
        # matches = len(glob(f"{path}/**/*.py", recursive=True)) <= THRESHOLD
        # but it would allegedly be slower since it needs to traverse the whole directory
        num_python_files = 0
        for root, dirs, files in os.walk(path):
            for file in files:
                if file.lower().endswith(".py"):
                    num_python_files += 1
                if num_python_files > THRESHOLD:
                    return False
        return True



""" Compromised Email Detector

Detects if a maintainer's email domain might have been compromised.
"""
from datetime import datetime
from typing import Optional

from dateutil import parser
from packaging import version

from cagged.analyzer.metadata.potentially_compromised_email_domain import \
    PotentiallyCompromisedEmailDomainDetector


class PypiPotentiallyCompromisedEmailDomainDetector(
    PotentiallyCompromisedEmailDomainDetector
):
    def __init__(self):
        super().__init__("pypi")

    def get_email_addresses(self, package_info: dict) -> list[str]:
        author_email = package_info["info"]["author_email"]
        maintainer_email = package_info["info"]["maintainer_email"]
        email = author_email or maintainer_email
        return [email]

    def get_project_latest_release_date(self, package_info) -> Optional[datetime]:
        """
        Gets the most recent release date of a Python project

        Args:
            releases (dict): PyPI JSON API's representation field

        Returns:
            datetime: creation date of the most recent in releases
        """
        releases = package_info["releases"]
        sorted_versions = sorted(
            releases.keys(), key=lambda r: version.parse(r), reverse=True
        )
        earlier_versions = (
            sorted_versions[:-1] if len(sorted_versions) > 1 else sorted_versions
        )

        for early_version in earlier_versions:
            version_release = releases[early_version]

            if len(version_release) > 0:  # if there's a distribution for the package
                upload_time_text = version_release[0]["upload_time_iso_8601"]
                release_date = parser.isoparse(upload_time_text).replace(tzinfo=None)
                return release_date
        raise Exception("could not find release date")



from abc import abstractmethod
from typing import Optional


class Detector:
    RULE_NAME = ""

    def __init__(self, name: str, description: str) -> None:
        self.name = name
        self.description = description

    @abstractmethod
    def detect(
        self,
        package_info,
        path: Optional[str] = None,
        name: Optional[str] = None,
        version: Optional[str] = None,
    ) -> tuple[bool, Optional[str]]:
        pass

    def get_name(self) -> str:
        return self.name

    def get_description(self) -> str:
        return self.description



""" PyPI Package Malware Scanner

CLI command that scans a PyPI package version for user-specified malware flags.
Includes rules based on package registry metadata and source code analysis.
"""
import logging
import os
import sys
from typing import Iterable, Optional, cast

import click
import rich
from pip._internal.cli.main import main as pip_main
from prettytable import PrettyTable
from rich.logging import RichHandler

from cagged.analyzer.analyzer import SEMGREP_RULE_NAMES
from cagged.analyzer.metadata import get_metadata_detectors
from cagged.analyzer.sourcecode import SOURCECODE_RULES
from cagged.dynamic.analyze import run_analysis
from cagged.ecosystems import ECOSYSTEM
from cagged.misc import console, console_without_time
from cagged.reporters.sarif import report_verify_sarif
from cagged.scanners import get_scanner
from cagged.scanners.scanner import PackageScanner
from cagged.utils.package_info import get_package_info

ALL_RULES = (
    set(get_metadata_detectors(ECOSYSTEM.NPM).keys())
    | set(get_metadata_detectors(ECOSYSTEM.PYPI).keys())
    | SEMGREP_RULE_NAMES
)
EXIT_CODE_ISSUES_FOUND = 1

AVAILABLE_LOG_LEVELS = {logging.DEBUG, logging.INFO, logging.WARN, logging.ERROR}
AVAILABLE_LOG_LEVELS_NAMES = list(
    map(lambda level: logging.getLevelName(level), AVAILABLE_LOG_LEVELS)
)

FORMAT = "%(message)s"
logging.basicConfig(
    level="INFO", format=FORMAT, datefmt="[%X]", handlers=[RichHandler(show_path=False)]
)

log = logging.getLogger("rich")


def common_options(fn):
    fn = click.option(
        "--exit-non-zero-on-finding",
        default=False,
        is_flag=True,
        help="Exit with a non-zero status code if at least one issue is identified",
    )(fn)
    fn = click.option(
        "-r",
        "--rules",
        multiple=True,
        type=click.Choice(ALL_RULES, case_sensitive=False),
    )(fn)
    fn = click.option(
        "-x",
        "--exclude-rules",
        multiple=True,
        type=click.Choice(ALL_RULES, case_sensitive=False),
    )(fn)
    fn = click.argument("target")(fn)
    return fn


def verify_options(fn):
    fn = click.option(
        "--output-format",
        default=None,
        type=click.Choice(["json", "sarif"], case_sensitive=False),
    )(fn)
    return fn


def scan_options(fn):
    fn = click.option(
        "--output-format",
        default=None,
        type=click.Choice(["json"], case_sensitive=False),
    )(fn)
    fn = click.option(
        "-v", "--version", default=None, help="Specify a version to scan"
    )(fn)
    return fn


def logging_options(fn):
    fn = click.option(
        "--log-level",
        default="INFO",
        type=click.Choice(AVAILABLE_LOG_LEVELS_NAMES, case_sensitive=False),
    )(fn)
    return fn


@click.group
@logging_options
def cli(log_level):
    """
    caged cli tool to detect malware in package ecosystems

    Supports PyPI and npm

    Example: caged pypi scan semantic-version

    Use --help for the detail of all commands and subcommands
    """
    logger = logging.getLogger("caged")
    logger.setLevel(logging.getLevelName(log_level))
    stdoutHandler = logging.StreamHandler(stream=sys.stdout)
    stdoutHandler.setFormatter(logging.Formatter("%(levelname)s: %(message)s"))
    logger.addHandler(stdoutHandler)
    pass


def _get_rule_pram(rules, exclude_rules):
    rule_param = None
    if len(rules) > 0:
        rule_param = rules
    if len(exclude_rules) > 0:
        rule_param = ALL_RULES - set(exclude_rules)
        if len(rules) > 0:
            print("--rules and --exclude-rules cannot be used together")
            exit(1)
    return rule_param


def _verify(
    path, rules, exclude_rules, output_format, exit_non_zero_on_finding, ecosystem
):
    """Verify a requirements.txt file

    Args:
        path (str): path to requirements.txt file
    """
    return_value = None
    rule_param = _get_rule_pram(rules, exclude_rules)
    scanner = get_scanner(ecosystem, True)
    if scanner is None:
        sys.stderr.write(f"Command verify is not supported for ecosystem {ecosystem}")
        exit(1)

    def display_result(result: dict) -> None:
        identifier = (
            result["dependency"]
            if result["version"] is None
            else f"{result['dependency']} version {result['version']}"
        )
        if output_format is None:
            print_scan_results(result.get("result"), identifier)

        if len(result.get("errors", [])) > 0:
            print_errors(result.get("error"), identifier)

    results = scanner.scan_local(path, rule_param, display_result)
    if output_format == "json":
        import json as js

        return_value = js.dumps(results)

    if output_format == "sarif":
        return_value = report_verify_sarif(path, list(ALL_RULES), results, ecosystem)

    if output_format is not None:
        print(return_value)

    if exit_non_zero_on_finding:
        exit_with_status_code(results)

    return return_value


def is_local_target(identifier: str) -> bool:
    """
    @param identifier:  The name/path of the package as passed to "caged ecosystem scan"
    @return:            Whether the identifier should be consider a local path
    """
    if (
        identifier.startswith("/")
        or identifier.startswith("./")
        or identifier.startswith("../")
    ):
        return True

    if identifier == ".":
        return True

    if (
        identifier.endswith(".tar.gz")
        or identifier.endswith(".zip")
        or identifier.endswith(".whl")
    ):
        return os.path.exists(identifier)

    return False


def is_verified_package(result: dict) -> bool:
    """
    @param result:  The result of a scan
    @return:        Whether the result is a verified package
    """
    return not result.get("issues")


def install_pypi_package(zip_path: str, pip_args: Iterable) -> None:
    """
    @param zip_path:    The path to the zip file
    @param pip_args:    Additional arguments to pass to pip
    """
    log.debug(f"Installing package from {zip_path}")
    pip_main(["install", zip_path, *pip_args])


def _scan(
    identifier,
    version,
    rules,
    exclude_rules,
    output_format,
    exit_non_zero_on_finding,
    ecosystem: ECOSYSTEM,
    install_package=False,
    pip_args=None,
):
    """Scan a package

    Args:
        identifier (str): name or path to the package
        version (str): version of the package (ex. 1.0.0), defaults to most recent
        rules (list[str]): specific rules to run, defaults to all
    """

    if not is_local_target(identifier):
        package_info = get_package_info(identifier)

        console_without_time.rule(f"[bold]Package information[/bold]")

        console_without_time.log(
            f"[bold][green]{identifier}[/green][/bold] – [italic]{package_info['info']['summary']}[/italic]",
            justify="center",
        )
        console_without_time.log(
            f"[bold]Version[/bold] – {package_info['info']['version']}",
            justify="center",
        )
        console_without_time.log(
            f"[bold]Author[/bold] – {package_info['info']['author']}", justify="center"
        )

        console.line()

    rule_param = _get_rule_pram(rules, exclude_rules)
    scanner = cast(Optional[PackageScanner], get_scanner(ecosystem, False))
    if scanner is None:
        sys.stderr.write(f"Command scan is not supported for ecosystem {ecosystem}")
        exit(1)

    results = {}
    if is_local_target(identifier):
        log.debug(
            f"Considering that '{identifier}' is a local target, scanning filesystem"
        )
        results, zip_path = scanner.scan_local(identifier, rule_param)
    else:
        log.debug(f"Considering that '{identifier}' is a remote target")
        try:
            results, zip_path = scanner.scan_remote(identifier, version, rule_param)
        except Exception as e:
            sys.stderr.write("\n")
            sys.stderr.write(str(e))
            sys.exit()

    if output_format == "json":
        import json as js

        print(js.dumps(results))
    else:
        print_scan_results(results, identifier)

    static_verdict = is_verified_package(results)

    if static_verdict:
        dynamic_verdict = not run_analysis(
            ecosystem="pypi",
            package=identifier,
            package_path=zip_path,
        )

        if install_package:
            if dynamic_verdict:
                console.log(
                    f"🟢 Considering that [bold]'{identifier}'[/bold] is a [green]valid[/green] package. [bold]Starting installation process[/bold]",
                    emoji=True,
                )
                print()
                install_pypi_package(zip_path, pip_args)
            else:
                console.log(
                    f"🔴 Considering that [bold]'{identifier}'[/bold] is a [red]malicious[/red] package. [bold]Installation [red]aborted[/red][/bold]",
                    emoji=True,
                )
    else:
        console.log(
            f"🔴 Considering that [bold]'{identifier}'[/bold] is a [red]malicious[/red] package. [bold]Installation [red]aborted[/red][/bold]",
            emoji=True,
        )

    if exit_non_zero_on_finding:
        exit_with_status_code(results)


def _list_rules(ecosystem):
    table = PrettyTable()
    table.align = "l"
    table.field_names = ["Rule type", "Rule name", "Description"]

    for rule in SOURCECODE_RULES[ecosystem]:
        table.add_row(
            ["Source code", rule["id"], rule.get("metadata", {}).get("description")]
        )

    metadata_rules = get_metadata_detectors(ecosystem)
    for ruleName in metadata_rules:
        rule = metadata_rules[ruleName]
        table.add_row(["Package metadata", rule.get_name(), rule.get_description()])

    print(table)


@cli.group
def npm(**kwargs):
    """Scan a npm package or verify a npm project"""
    pass


@cli.group
def pip(**kwargs):
    """Scan a PyPI package or verify a PyPI project"""
    pass


@npm.command("scan")
@common_options
@scan_options
def scan_npm(
    target, version, rules, exclude_rules, output_format, exit_non_zero_on_finding
):
    """Scan a given npm package"""
    return _scan(
        target,
        version,
        rules,
        exclude_rules,
        output_format,
        exit_non_zero_on_finding,
        ECOSYSTEM.NPM,
    )


@npm.command("verify")
@common_options
@verify_options
def verify_npm(target, rules, exclude_rules, output_format, exit_non_zero_on_finding):
    """Verify a given npm project"""
    return _verify(
        target,
        rules,
        exclude_rules,
        output_format,
        exit_non_zero_on_finding,
        ECOSYSTEM.NPM,
    )


@pip.command(
    "install",
    context_settings=dict(
        ignore_unknown_options=True,
    ),
)
@common_options
@scan_options
@click.argument("pip_args", nargs=-1, type=click.UNPROCESSED)
def install_pypi(
    target,
    version,
    rules,
    exclude_rules,
    output_format,
    exit_non_zero_on_finding,
    pip_args,
):
    """Scan and install a given PyPI package"""
    return _scan(
        target,
        version,
        rules,
        exclude_rules,
        output_format,
        exit_non_zero_on_finding,
        ECOSYSTEM.PYPI,
        install_package=True,
        pip_args=pip_args,
    )


@pip.command("scan")
@common_options
@scan_options
def scan_pypi(
    target, version, rules, exclude_rules, output_format, exit_non_zero_on_finding
):
    """Scan a given PyPI package"""
    return _scan(
        target,
        version,
        rules,
        exclude_rules,
        output_format,
        exit_non_zero_on_finding,
        ECOSYSTEM.PYPI,
    )


@pip.command("verify")
@common_options
@verify_options
def verify_pypi(target, rules, exclude_rules, output_format, exit_non_zero_on_finding):
    """Verify a given Pypi project"""
    return _verify(
        target,
        rules,
        exclude_rules,
        output_format,
        exit_non_zero_on_finding,
        ECOSYSTEM.PYPI,
    )


@pip.command("list-rules")
def list_rules_pypi():
    """Print available rules for PyPI"""
    return _list_rules(ECOSYSTEM.PYPI)


@npm.command("list-rules")
def list_rules_npm():
    """Print available rules for npm"""
    return _list_rules(ECOSYSTEM.NPM)


@cli.command("verify", deprecated=True)
@common_options
@verify_options
def verify(target, rules, exclude_rules, output_format, exit_non_zero_on_finding):
    return _verify(
        target,
        rules,
        exclude_rules,
        output_format,
        exit_non_zero_on_finding,
        ECOSYSTEM.PYPI,
    )


@cli.command("scan", deprecated=True)
@common_options
@scan_options
def scan(
    target, version, rules, exclude_rules, output_format, exit_non_zero_on_finding
):
    return _scan(
        target,
        version,
        rules,
        exclude_rules,
        output_format,
        exit_non_zero_on_finding,
        ECOSYSTEM.PYPI,
    )


def print_scan_results(results, identifier):
    num_issues = results.get("issues")
    errors = results.get("errors", [])

    if num_issues == 0:
        console.log(
            f"🟢 Found [green]0[/green] potentially malicious indicators scanning [bold][green]{identifier}[/green][/bold]",
            emoji=True,
        )
        console.line()
    else:
        console.log(
            f"🔴 Found [red]{num_issues}[/red] potentially malicious indicators scanning [bold][green]{identifier}[/green][/bold]",
            emoji=True,
        )

        findings = results.get("results", [])
        for finding in findings:
            description = findings[finding]
            if type(description) == str:
                console.log(f"[bold]{finding}[/bold]: {description}")
                console.line()
            elif type(description) == list:
                source_code_findings = description
                console.log(
                    f"[bold][red]{finding}[/red][/bold]: found [red]{len(source_code_findings)}[/red] source code matches"
                )
                for finding in source_code_findings:
                    console.log(
                        f"  * {finding['message']} at {finding['location']} \n      {format_code_line_for_output(finding['code'])}"
                    )
                console.line()
                print()

    if len(errors) > 0:
        print_errors(errors, identifier)
        print("\n")


def print_errors(errors, identifier):
    console.log(
        f"[yellow]Some rules failed to run while scanning [bold][green]{identifier}[/green][/bold]:[/yellow]"
    )
    console.line()
    for rule in errors:
        console.log(
            f"  * [bold][yellow]{rule}[/yellow][/bold]: [italic]{errors[rule]}[/italic]"
        )
    console.line()


def format_code_line_for_output(code):
    return "    " + code.strip().replace("\n", "\n    ").replace("\t", "  ")


def exit_with_status_code(results):
    num_issues = results.get("issues", 0)
    if num_issues > 0:
        exit(EXIT_CODE_ISSUES_FOUND)



""" Empty Information Detector

Detects when a package has its latest release version to 0.0.0
"""
from typing import Optional

from cagged.analyzer.metadata.release_zero import ReleaseZeroDetector


class NPMReleaseZeroDetector(ReleaseZeroDetector):
    def detect(
        self,
        package_info,
        path: Optional[str] = None,
        name: Optional[str] = None,
        version: Optional[str] = None,
    ) -> tuple[bool, str]:
        return (
            package_info["dist-tags"]["latest"] in ["0.0.0", "0.0", "0"],
            ReleaseZeroDetector.MESSAGE_TEMPLATE % package_info["dist-tags"]["latest"],
        )



import json
import logging
import os
import subprocess
from pathlib import Path
from typing import Iterable, List, Optional

import rich
from rich.progress import track
from rich.table import Column, Table

from cagged.analyzer.metadata import get_metadata_detectors
from cagged.ecosystems import ECOSYSTEM


def get_rules(file_extension, path):
    return set(
        rule.replace(file_extension, "")
        for rule in os.listdir(path)
        if rule.endswith(file_extension)
    )


SEMGREP_RULES_PATH = os.path.join(os.path.dirname(__file__), "sourcecode")
SEMGREP_RULE_NAMES = get_rules(".yml", SEMGREP_RULES_PATH)

log = logging.getLogger("caged")


class Analyzer:
    """
    Analyzes a local directory for threats found by source code or metadata rules

    Attributes:
        sourcecode_rules_path (str): path to source code rules
        ecosystem (str): name of the current ecosystem
        metadata_ruleset (list): list of metadata rule names
        sourcecode_ruleset (list): list of source code rule names

        exclude (list): list of directories to exclude from source code search

        metadata_detectors(list): list of metadata detectors
    """

    def __init__(self, ecosystem=ECOSYSTEM.PYPI) -> None:
        self.sourcecode_rules_path = os.path.join(
            os.path.dirname(__file__), "sourcecode"
        )

        self.ecosystem = ecosystem

        self.metadata_detectors = get_metadata_detectors(ecosystem)

        self.metadata_ruleset = self.metadata_detectors.keys()
        self.sourcecode_ruleset = SEMGREP_RULE_NAMES

        self.exclude = [
            "helm",
            ".idea",
            "venv",
            "test",
            "tests",
            ".env",
            "dist",
            "build",
            "semgrep",
            "migrations",
            ".github",
            ".semgrep_logs",
        ]

    def analyze(
        self,
        path,
        info=None,
        rules=None,
        name: Optional[str] = None,
        version: Optional[str] = None,
    ) -> dict:
        """
        Analyzes a package in the given path

        Args:
            path (str): path to package
            info (dict, optional): Any package information to analyze metadata. Defaults to None.
            rules (set, optional): Set of rules to analyze. Defaults to all rules.

        Raises:
            Exception: "{rule} is not a valid rule."

        Returns:
            dict[str]: map from each rule and their corresponding output
        """

        metadata_results = None
        sourcecode_results = None

        metadata_rules = None
        sourcecode_rules = None
        if rules is not None:

            sourcecode_rules = set()
            metadata_rules = set()

            for rule in rules:
                if rule in self.sourcecode_ruleset:
                    log.debug(f"Using source code rule {rule}")
                    sourcecode_rules.add(rule)
                elif rule in self.metadata_ruleset:
                    log.debug(f"Using metadata rule {rule}")
                    metadata_rules.add(rule)
                else:
                    raise Exception(f"{rule} is not a valid rule.")

        log.debug(f"Running metadata rules against package '{name}'")
        metadata_results = self.analyze_metadata(
            path, info, metadata_rules, name, version
        )

        log.debug(f"Running source code rules against directory '{path}'")
        sourcecode_results = self.analyze_sourcecode(path, sourcecode_rules)

        issues = metadata_results["issues"] + sourcecode_results["issues"]
        results = metadata_results["results"] | sourcecode_results["results"]
        errors = metadata_results["errors"] | sourcecode_results["errors"]

        return {"issues": issues, "errors": errors, "results": results, "path": path}

    def analyze_metadata(
        self,
        path: str,
        info,
        rules=None,
        name: Optional[str] = None,
        version: Optional[str] = None,
    ) -> dict:
        """
        Analyzes the metadata of a given package

        Args:
            path (str): path to package
            info (dict): package information given by PyPI Json API
            rules (set, optional): Set of metadata rules to analyze. Defaults to all rules.

        Returns:
            dict[str]: map from each metadata rule and their corresponding output
        """

        all_rules = rules if rules is not None else self.metadata_ruleset
        results = {}
        errors = {}
        issues = 0

        table = Table(
            *all_rules,
            title="Metadata analysis",
        )

        results_line = []

        for rule in all_rules:
            try:
                log.debug(f"Running rule {rule} against package '{name}'")
                rule_matches, message = self.metadata_detectors[rule].detect(
                    info, path, name, version
                )
                if rule_matches:
                    issues += 1
                    results[rule] = message
                    results_line.append("[red]FOUND[/red]")
                else:
                    results_line.append("[green]OK[/green]")
            except Exception as e:
                errors[rule] = f"failed to run rule {rule}: {str(e)}"

        table.add_row(*results_line)

        rich.print(table)
        print()

        return {"results": results, "errors": errors, "issues": issues}

    def analyze_sourcecode(self, path, rules=None) -> dict:
        """
        Analyzes the source code of a given package

        Args:
            path (str): path to directory of package
            rules (set, optional): Set of source code rules to analyze. Defaults to all rules.

        Returns:
            dict[str]: map from each source code rule and their corresponding output
        """
        targetpath = Path(path)
        all_rules = rules if rules is not None else self.sourcecode_ruleset
        results = {rule: {} for rule in all_rules}  # type: dict
        errors = {}
        issues = 0

        rules_path: List[str]
        if rules is None:
            log.debug(
                f"No rules specified using full rules directory {self.sourcecode_rules_path}"
            )
            rules_path = [self.sourcecode_rules_path]
        else:
            rules_path = list(
                map(
                    lambda rule_name: os.path.join(
                        self.sourcecode_rules_path, f"{rule_name}.yml"
                    ),
                    rules,
                )
            )

        all_rules_list = list(all_rules)

        table = Table(
            *all_rules_list,
            title="Source code analysis",
        )

        results_line = ["[green]OK[/green]" for _ in all_rules_list]

        if len(rules_path) == 0:
            log.debug("No source code rules to run")
            return {"results": {}, "errors": {}, "issues": 0}

        try:
            log.debug(f"Running source code rules against {path}")
            response = self._invoke_semgrep(target=path, rules=rules_path)
            rule_results = self._format_semgrep_response(
                response, targetpath=targetpath
            )
            for found_rule in rule_results:
                results_line[all_rules_list.index(found_rule)] = "[red]FOUND[/red]"
            issues += len(rule_results)

            results = results | rule_results
        except Exception as e:
            errors["rules-all"] = f"failed to run rule: {str(e)}"

        table.add_row(*results_line)
        rich.print(table)
        print()

        return {"results": results, "errors": errors, "issues": issues}

    def _invoke_semgrep(self, target: str, rules: Iterable[str]):
        try:
            cmd = ["semgrep"]
            for rule in rules:
                cmd.extend(["--config", rule])

            for excluded in self.exclude:
                cmd.append(f"--exclude='{excluded}'")
            cmd.append("--no-git-ignore")
            cmd.append("--json")
            cmd.append("--quiet")
            cmd.append(target)
            log.debug(f"Invoking semgrep with command line: {' '.join(cmd)}")
            result = subprocess.run(
                cmd, capture_output=True, check=True, encoding="utf-8"
            )
            return json.loads(str(result.stdout))
        except FileNotFoundError:
            raise Exception("unable to find semgrep binary")
        except subprocess.CalledProcessError as e:
            error_message = f"""
An error occurred when running Semgrep.

command: {" ".join(e.cmd)}
status code: {e.returncode}
output: {e.output}
"""
            raise Exception(error_message)
        except json.JSONDecodeError as e:
            raise Exception("unable to parse semgrep JSON output: " + str(e))

    def _format_semgrep_response(self, response, rule=None, targetpath=None):
        """
        Formats the response from Semgrep

        Args:
            response (dict): response from Semgrep
            rule (str, optional): name of rule to format. Defaults to all rules.
            targetpath (str, optional): root directory of scan. Defaults to None.
                Paths in formatted resonse will be rooted from targetpath.

        Returns:
            dict: formatted response in the form...

            {
                ...
                <rule-name>: {
                    <path-to-code:line-num>: <dangerous-code>
                    ...
                },
                ...
            }
        """

        results = {}

        for result in response["results"]:
            rule_name = rule or result["check_id"].split(".")[-1]
            code_snippet = result["extra"]["lines"]
            line = result["start"]["line"]

            file_path = os.path.abspath(result["path"])
            if targetpath:
                file_path = os.path.relpath(file_path, targetpath)

            location = file_path + ":" + str(line)
            code = self.trim_code_snippet(code_snippet)

            if rule_name not in result:
                results[rule_name] = []
                results[rule_name].append(
                    {
                        "location": location,
                        "code": code,
                        "message": result["extra"]["message"],
                    }
                )

        return results

    # Makes sure the matching code to be displayed isn't too long
    def trim_code_snippet(self, code):
        THRESHOLD = 250
        if len(code) > THRESHOLD:
            return code[: THRESHOLD - 10] + "..." + code[len(code) - 10 :]
        else:
            return code
